/* kruskal.c */

/* Direttive */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>


#define ROW_LENGTH 256

#include "grafo-la.h"
#include "alloc.h"


typedef nodo * vnodo;
typedef posarco * vposarco;

typedef struct _soluzioneMSTP soluzioneMSTP;
struct _soluzioneMSTP
{
  int f;
  int dim;
  int nl;
  vnodo v1;
  vnodo v2;
};

/* Prototipi delle procedure secondarie */

void InterpretaLineaComando (int argc, char *argv[], char *filedati);

/* Carica da filedati il grafo G */
void CaricaDati (char *filedati, grafo *pG);

/* Crea una soluzione vuota */
void CreaSoluzione (int n, soluzioneMSTP *pS);

/* Distrugge la soluzione S */
void DistruggeSoluzione (soluzioneMSTP *pS);

/* Stampa la soluzione */
void StampaSoluzione (soluzioneMSTP *pS);

/* Risolve l'istanza G del MSTP con l'algoritmo di Kruskal restituendo la soluzione S */
void Kruskal (grafo *pG, soluzioneMSTP *pS);


/* Programma principale */
int main (int argc, char *argv[])
{
  /* Parte dichiarativa */
  char filedati[ROW_LENGTH];
  grafo G;
  soluzioneMSTP S;
  clock_t inizio, fine;
  double tempo;


  /* Parte esecutiva */
  InterpretaLineaComando(argc,argv,filedati);

  /* Carica i dati */
  CaricaDati(filedati,&G);

  /* Crea una soluzione vuota */
  CreaSoluzione(G.n-1,&S);

  /* Risolve il problema dell'albero ricoprente minimo sul grafo G */
  inizio = clock();
  Kruskal(&G,&S);
  fine = clock();
  tempo = ((double) (fine - inizio))/CLOCKS_PER_SEC;

  /* Stampa il tempo di calcolo e la soluzione */
  printf("CPU = %f\n",tempo);
  StampaSoluzione(&S);

  /* Dealloca le strutture dati dinamiche */
  distruggegrafo(&G);
  DistruggeSoluzione(&S);

  return EXIT_SUCCESS;
}


/* Definizione delle procedure secondarie */

void InterpretaLineaComando (int argc, char *argv[], char *filedati)
{
  if (argc != 2)
  {
    fprintf(stderr,"Errore nella linea di comando!\n");
    fprintf(stderr,"%s [filedati]\n",argv[0]);
    exit(EXIT_FAILURE);
  }

  strcpy(filedati,argv[1]);
}


/* Carica da filedati il grafo G */
void CaricaDati (char *filedati, grafo *pG)
{
  FILE *fp;
  int n, m, e;
  nodo u, v;
  int c;


  fp = fopen(filedati,"r");
  if (fp == NULL)
  {
    fprintf(stderr,"Errore nell'apertura del file %s!\n",filedati);
    exit(EXIT_FAILURE);
  }

  if (fscanf(fp,"%*s %*s %d",&n) != 1)
  {
    fprintf(stderr,"Errore nella lettura del numero di vertici!\n");
    exit(EXIT_FAILURE);
  }

  creagrafo(n, pG);

  if (fscanf(fp,"%*s %*s %d",&m) != 1)
  {
    fprintf(stderr,"Errore nella lettura del numero di lati!\n");
    exit(EXIT_FAILURE);
  }

  for (e = 1; e <= m; e++)
  {
    if (fscanf(fp," (%d ,%d ) %d",&u,&v,&c) != 3)
    {
      fprintf(stderr,"Errore nella lettura di un lato!\n");
      exit(EXIT_FAILURE);
    }
    insarco(u, v, c, pG);
  }

  fclose(fp);
}


/* Crea una soluzione vuota */
void CreaSoluzione (int n, soluzioneMSTP *pS)
{
  pS->f = 0;
  pS->dim = n;
  pS->nl = 0;
  pS->v1 = (vint) calloc(pS->dim+1,sizeof(int));
  if (pS->v1 == NULL)
  {
    fprintf(stderr,"Errore nell'allocazione del vettore dei primi estremi!\n");
    exit(EXIT_FAILURE);
  }
  pS->v2 = (vint) calloc(pS->dim+1,sizeof(int));
  if (pS->v2 == NULL)
  {
    fprintf(stderr,"Errore nell'allocazione del vettore dei secondi estremi!\n");
    exit(EXIT_FAILURE);
  }
}


/* Distrugge la soluzione S */
void DistruggeSoluzione (soluzioneMSTP *pS)
{
  pS->f = 0;
  pS->dim = pS->nl = 0;
  free(pS->v1);
  pS->v1 = NULL;
  free(pS->v2);
  pS->v2 = NULL;
}


/* Stampa la soluzione */
void StampaSoluzione (soluzioneMSTP *pS)
{
  int i;

  printf("f* = %d\n",pS->f);
  printf("x* = ");
  for (i = 1; i <= pS->nl; i++)
    printf(" (%d,%d)",pS->v1[i],pS->v2[i]);
  printf("\n");
}


/* Risolve l'istanza G del MSTP con l'algoritmo di Kruskal restituendo la soluzione S */
void Kruskal (grafo *pG, soluzioneMSTP *pS)
{
}
